@@ -31,8 +31,13 @@ class Speaker(EnumByName):
 class DialogueFlow:
 
     Speaker = Speaker
+    _autostate = '-1'
+    @classmethod
+    def autostate(cls):
+        cls._autostate = str(int(cls._autostate) + 1)
+        return cls._autostate
 
-    def __init__(self, initial_state: Union[Enum, str], initial_speaker = Speaker.SYSTEM,
+    def __init__(self, initial_state: Union[Enum, str, tuple], initial_speaker = Speaker.SYSTEM,
                  macros: Dict[str, Macro] =None, kb: Union[KnowledgeBase, str, List[str]] =None):
         self._graph = Graph()
         self._initial_state = State(initial_state)
@@ -46,6 +51,7 @@ class DialogueFlow:
         self._gate_buffer = {}
         self._var_dependencies = defaultdict(set)
         self._error_transitioned = False
+        self._prepends = {}
         if kb is None:
             self._kb = KnowledgeBase()
         elif isinstance(kb, str):
@@ -79,7 +85,9 @@ class DialogueFlow:
             'PSP': PossessivePronoun(self._kb),
             'EQ': Equal(),
             'GATE': Gate(self),
-            'CLR': Clear()
+            'CLR': Clear(),
+            'NER': NamedEntity(),
+            'POS': PartOfSpeech()
         }
         if macros:
             self._macros.update(macros)
@@ -155,9 +163,94 @@ class DialogueFlow:
         if debugging:
             print('User turn in {:.5f}'.format(t2 - t1))
 
+
+    def load_transitions(self, json_dict, speaker=None, root=None):
+        """
+        wheeeeeeee!
+        """
+        if speaker is None:
+            speaker = self._initial_speaker
+        if 'state' in json_dict:
+            source = json_dict['state']
+        else:
+            source = DialogueFlow.autostate()
+        if root is None:
+            root = self._initial_state
+
+        hop = None
+
+        # read settings and transitions for state
+        transitions = []
+        for key, value in json_dict.items():
+            if key == 'transitions':
+                assert isinstance(value, list)
+                transitions = value
+            elif key == 'root':
+                root = json_dict['root']
+            elif key == 'hop':
+                hop = json_dict['hop']
+            elif key == 'prepend':
+                prepend = json_dict['prepend']
+                self.set_state_prepend(source, prepend)
+            elif key not in {'state', 'hop', 'score'}:
+                transitions.append((key, value))
+
+        # set up state settings
+        if not self.has_state(source):
+            self.add_state(source)
+            self.set_error_successor(source, root)
+        if hop:
+            if speaker == Speaker.USER:
+                speaker = Speaker.SYSTEM
+                self.state_settings(source).update(system_multi_hop=True)
+            elif speaker == Speaker.SYSTEM:
+                speaker = Speaker.USER
+                self.state_settings(source).update(user_multi_hop=True)
+
+        # set up transitions
+        expanded_transitions = []
+        for natex, target in transitions:
+            if natex == 'error':
+                if isinstance(target, dict):
+                    if 'state' not in target:
+                        target['state'] = DialogueFlow.autostate()
+                    expanded_transitions.append(target)
+                    target = target['state']
+                    if not self.has_state(target):
+                        self.add_state(target)
+                        self.set_error_successor(target, root)
+                self.set_error_successor(source, target)
+
+            else:
+                score = 1.0
+                if isinstance(target, dict):
+                    if 'state' not in target:
+                        target['state'] = DialogueFlow.autostate()
+                    if 'score' in target:
+                        score = target['score']
+                    expanded_transitions.append(target)
+                    target = target['state']
+                    if not self.has_state(target):
+                        self.add_state(target)
+                        self.set_error_successor(target, root)
+                if speaker == Speaker.USER:
+                    self.add_user_transition(source, target, natex, score=score)
+                elif speaker == Speaker.SYSTEM:
+                    self.add_system_transition(source, target, natex, score=score)
+
+        # swtich turn (will be switched back if multi hop detected on next recursive call)
+        if speaker == Speaker.USER:
+            speaker = Speaker.SYSTEM
+        elif speaker == Speaker.SYSTEM:
+            speaker = Speaker.USER
+
+        # recurse to load nested turns
+        for transition in expanded_transitions:
+            self.load_transitions(transition, speaker)
+
     # HIGH LEVEL
 
-    def system_transition(self, state: Union[Enum, str], debugging=False):
+    def system_transition(self, state: Union[Enum, str, tuple], debugging=False):
         """
         :param state:
         :param debugging:
@@ -177,7 +270,14 @@ class DialogueFlow:
             natex = self.transition_natex(*transition)
             settings = self.transition_settings(*transition)
             vars = HashableDict(self._vars)
-            generation = natex.generate(vars=vars, macros=self._macros, debugging=debugging)
+            try:
+                generation = natex.generate(vars=vars, macros=self._macros, debugging=debugging)
+            except Exception as e:
+                print()
+                print(e)
+                print('Transition {}: {} failed'.format(str(transition), natex))
+                print()
+                generation = None
             if generation is not None:
                 transition_options[(generation, transition, vars)] = settings.score
             t2 = time()
@@ -196,8 +296,7 @@ class DialogueFlow:
                         del transition_options[key]
                 else:
                     break
-            options = StochasticOptions(transition_options)
-            response, transition, vars = options.select()
+            response, transition, vars = max(transition_options, key=lambda x: transition_options[x])
             if debugging:
                 updates = {}
                 for k, v in vars.items():
@@ -222,7 +321,7 @@ class DialogueFlow:
         else:
             raise AssertionError('dialogue flow system transition found no valid options')
 
-    def user_transition(self, natural_language: str, state: Union[Enum, str], debugging=False):
+    def user_transition(self, natural_language: str, state: Union[Enum, str, tuple], debugging=False):
         """
         :param state:
         :param natural_language:
@@ -247,9 +346,17 @@ class DialogueFlow:
             natex = self.transition_natex(*transition)
             settings = self.transition_settings(*transition)
             vars = HashableDict(self._vars)
-            match = natex.match(natural_language, vars, self._macros, ngrams, debugging)
+            try:
+                match = natex.match(natural_language, vars, self._macros, ngrams, debugging)
+            except Exception as e:
+                print()
+                print(e)
+                print('Transition {}: {} failed'.format(str(transition), natex))
+                print()
+                match = None
             if match:
-                print('Transition {} matched "{}"'.format(transition[:2], natural_language))
+                if debugging:
+                    print('Transition {} matched "{}"'.format(transition[:2], natural_language))
                 transition_options.append((settings.score, transition, vars))
             t2 = time()
             if debugging:
@@ -284,6 +391,14 @@ class DialogueFlow:
                 print('Error transition {} -> {}'.format(self.state(), next_state))
             return next_state
 
+    def precache_transitions(self):
+        """
+        Make DialogueFlow fast from the start with the power of precache!
+        """
+        for transition in self._graph.arcs():
+            data = self._graph.arc_data(*transition)
+            data['natex'].precache()
+
     def check(self, debugging=False):
         all_good = True
         for state in self._graph.nodes():
@@ -308,7 +423,7 @@ class DialogueFlow:
                     all_good = False
         return all_good
 
-    def add_user_transition(self, source: Union[Enum, str], target: Union[Enum, str],
+    def add_user_transition(self, source: Union[Enum, str, tuple], target: Union[Enum, str, tuple],
                             natex_nlu: Union[str, NatexNLU, List[str]], **settings):
         source = State(source)
         target = State(target)
@@ -327,8 +442,12 @@ class DialogueFlow:
         transition_settings = Settings(score=1.0)
         transition_settings.update(**settings)
         self.set_transition_settings(source, target, Speaker.USER, transition_settings)
+        if target in self._prepends:
+            prepend = self._prepends[target]
+            natex = self.transition_natex(source, target, Speaker.USER)
+            self.set_transition_natex(source, target, Speaker.USER, prepend + natex)
 
-    def add_system_transition(self, source: Union[Enum, str], target: Union[Enum, str],
+    def add_system_transition(self, source: Union[Enum, str, tuple], target: Union[Enum, str, tuple],
                               natex_nlg: Union[str, NatexNLG, List[str]], **settings):
         source = State(source)
         target = State(target)
@@ -344,8 +463,12 @@ class DialogueFlow:
         transition_settings = Settings(score=1.0)
         transition_settings.update(**settings)
         self.set_transition_settings(source, target, Speaker.SYSTEM, transition_settings)
+        if target in self._prepends:
+            prepend = self._prepends[target]
+            natex = self.transition_natex(source, target, Speaker.SYSTEM)
+            self.set_transition_natex(source, target, Speaker.SYSTEM, prepend + natex)
 
-    def add_state(self, state: Union[Enum, str], error_successor: Union[Union[Enum, str], None] =None, **settings):
+    def add_state(self, state: Union[Enum, str, tuple], error_successor: Union[Union[Enum, str, tuple], None] =None, **settings):
         state = State(state)
         if self.has_state(state):
             raise ValueError('state {} already exists'.format(state))
@@ -376,7 +499,7 @@ class DialogueFlow:
 
     # LOW LEVEL: PROPERTIES, GETTERS, SETTERS
 
-    def transition_natex(self, source: Union[Enum, str], target: Union[Enum, str], speaker: Enum):
+    def transition_natex(self, source: Union[Enum, str, tuple], target: Union[Enum, str, tuple], speaker: Enum):
         source = State(source)
         target = State(target)
         return self._graph.arc_data(source, target, speaker)['natex']
@@ -386,7 +509,7 @@ class DialogueFlow:
         target = State(target)
         self._graph.arc_data(source, target, speaker)['natex'] = natex
 
-    def transition_settings(self, source: Union[Enum, str], target: Union[Enum, str], speaker: Enum):
+    def transition_settings(self, source: Union[Enum, str, tuple], target: Union[Enum, str, tuple], speaker: Enum):
         source = State(source)
         target = State(target)
         return self._graph.arc_data(source, target, speaker)['settings']
@@ -401,7 +524,7 @@ class DialogueFlow:
         target = State(target)
         self.transition_settings(source, target, speaker).update(**settings)
 
-    def transition_global_nlu(self, source: Union[Enum, str], target: Union[Enum, str]):
+    def transition_global_nlu(self, source: Union[Enum, str, tuple], target: Union[Enum, str, tuple]):
         source = State(source)
         target = State(target)
         if self.has_transition(source, target, Speaker.USER) \
@@ -415,7 +538,7 @@ class DialogueFlow:
         self._graph.arc_data(source, target, Speaker.USER)['global'] = is_global
         self.update_transition_settings(source, target, Speaker.USER, score=0.5)
 
-    def state_settings(self, state: Enum):
+    def state_settings(self, state):
         state = State(state)
         return self._graph.data(state)['settings']
 
@@ -456,7 +579,7 @@ class DialogueFlow:
     def state(self):
         return self._state
 
-    def set_state(self, state: Union[Enum, str]):
+    def set_state(self, state: Union[Enum, str, tuple]):
         state = State(state)
         self._state = state
 
@@ -527,16 +650,20 @@ class DialogueFlow:
         self._state = self._initial_state
         self._speaker = self._initial_speaker
         self._vars = {}
+        self._gates = defaultdict(set)
         for state in self.graph().nodes():
             self.state_settings(state).memory.clear()
 
     def update_vars(self, vars: HashableDict):
+        if not isinstance(vars, HashableDict):
+            vars = HashableDict(vars)
         for k in vars.altered():
             if k in self._var_dependencies:
                 dependencies = self._var_dependencies[k]
                 for dependency in dependencies:
                     if dependency in self._vars:
                         self._vars[dependency] = None
+
         self._vars.update({k: vars[k] for k in vars.altered()})
 
     def potential_transition(self):
@@ -560,6 +687,13 @@ class DialogueFlow:
     def var_dependencies(self):
         return self._var_dependencies
 
+    def set_state_prepend(self, state, prepend):
+        self._prepends[state] = prepend
+        if self.has_state(state):
+            for transition in self._graph.arcs_in(state):
+                natex = self.transition_natex(*transition)
+                self.set_transition_natex(*transition, prepend + natex)
+
     def passes_gate(self, var_config):
         if var_config in self._gates[self._potential_transition]:
             return False
