@@ -3,7 +3,7 @@ import regex
 from lark import Lark, Transformer, Tree, Visitor, Token
 from emora_stdm.state_transition_dialogue_manager.ngrams import Ngrams
 from emora_stdm.state_transition_dialogue_manager.utilities import HashableDict
-
+from copy import deepcopy
 
 class NatexNLU:
 
@@ -24,6 +24,7 @@ class NatexNLU:
             self._expression = expression.expression()
             self._macros = dict(expression.macros())
             self._macros.update(macros)
+        self._compiler = NatexNLU.Compiler(self._expression)
 
     def match(self, natural_language, vars=None, macros=None, ngrams=None, debugging=False):
         natural_language += ' _END_'
@@ -56,7 +57,10 @@ class NatexNLU:
             print('  {:15} {}'.format('Vars', ', '.join([k + '=' + str(v) for k, v in vars.items()])))
             print('  {:15} {}'.format('Steps', '  ' + '-' * 60))
             print('    {:15} {}'.format('Original', self._expression))
-        self._regex = NatexNLU.Compiler(ngrams, vars, macros, debugging).compile(self._expression)
+        self._regex = self._compiler.compile(ngrams, vars, macros, debugging)
+
+    def precache(self):
+        self._compiler.parse()
 
     def regex(self):
         return self._regex
@@ -70,8 +74,22 @@ class NatexNLU:
     def set_macros(self, macros):
         self._macros = macros
 
+    def __add__(self, other):
+        if isinstance(other, str):
+            return NatexNLU('[!' + self._expression + ', ' + other + ']', macros=self._macros)
+        elif isinstance(other, NatexNLU):
+            return NatexNLU('[!' + self._expression + ', ' + other._expression + ']', macros=self._macros)
+        return self
+
+    def __radd__(self, other):
+        if isinstance(other, str):
+            return NatexNLU('[!' + other + ', ' + self._expression + ']', macros=self._macros)
+        elif isinstance(other, NatexNLU):
+            return NatexNLU('[!' + other._expression + ', ' + self._expression + ']', macros=self._macros)
+        return self
+
     def __str__(self):
-        return 'Natex({})'.format(self._expression)
+        return 'NatexNLU({})'.format(self._expression)
 
     def __repr__(self):
         return str(self)
@@ -92,24 +110,37 @@ class NatexNLU:
         regex: "/" regex_value "/"
         reference: "$" symbol
         assignment: "$" symbol "=" term
-        macro: "#" symbol ( "(" term? (","? " "? term)* ")" )? 
+        macro: "#" symbol ( "(" macro_arg? (","? " "? macro_arg)* ")" )? 
+        macro_arg: macro_literal | macro
+        macro_literal: /[^#), ][^#),]*/
         literal: /[a-z_A-Z@.:]+( +[a-z_A-Z@.:]+)*/ | "\"" /[^\"]+/ "\""
         symbol: /[a-z_A-Z.0-9]+/
         regex_value: /[^\/]+/
         """
         parser = Lark(grammar)
 
-        def __init__(self, ngrams, vars, macros, debugging=False):
-            self._tree = None
+        def __init__(self, natex):
+            self._natex = natex
+            self._parsed_tree = None
+            self._tree = deepcopy(self._parsed_tree)
+            self._ngrams = None
+            self._vars = None
+            self._macros = None
+            self._assignments = set()
+            self._debugging = False
+
+        def parse(self):
+            self._parsed_tree = self.parser.parse(self._natex)
+
+        def compile(self, ngrams, vars, macros, debugging=False):
+            if self._parsed_tree is None:
+                self.parse()
+            self._tree = deepcopy(self._parsed_tree)
+            self._assignments = set()
             self._ngrams = ngrams
             self._vars = vars
             self._macros = macros
-            self._assignments = set()
             self._debugging = debugging
-            self._previous_compile_output = ''
-
-        def compile(self, natex):
-            self._tree = self.parser.parse(natex)
             re = self.visit(self._tree).children[0]
             if self._debugging:
                 print('  {:15} {}'.format('Final', re))
@@ -232,6 +263,13 @@ class NatexNLU:
                 if self._debugging: print('ERROR: Macro {} not found'.format(symbol))
                 tree.children[0] = '_MACRO_NOT_FOUND_'
 
+        def macro_arg(self, tree):
+            tree.children[0] = tree.children[0].children[0]
+            tree.data = 'compiled'
+
+        def macro_literal(self, tree):
+            tree.data = 'compiled'
+
         def literal(self, tree):
             args = tree.children
             tree.data = 'compiled'
@@ -285,6 +323,10 @@ class NatexNLU:
                     return '#' + args[0] + '(' + ', '.join([str(arg) for arg in args[1:]]) + ')'
                 def literal(self, args):
                     return str(args[0])
+                def macro_arg(self, args):
+                    return str(args[0])
+                def macro_literal(self, args):
+                    return str(args[0])
                 def symbol(self, args):
                     return str(args[0])
                 def term(self, args):
